/*
 *  CDEmu kernel module
 *  Copyright (C) 2006-2007 Rok Mandeljc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <linux/version.h>
#include <linux/autoconf.h>

#include <linux/poll.h>
#include <linux/cdev.h>
#include <linux/blkdev.h>
#include <linux/cdrom.h>


/* CDEmu driver structure */
struct cdemu_driver {
    char *version;
    
    struct module *module;
    struct device_driver driver;
    
    /* Module parameters */
    int ctl_major;
    int block_major;
    int devices;
    
    /* Control devices class */
    struct class *cdemu_ctl_class;
};
#define to_cdemu_driver(drv) container_of(drv, struct cdemu_driver, driver);


/* CDEmu device structure */
struct cdemu_device {
    int number; /* Device number */
    char name[32]; /* Device name */
    int debug_mask; /* Debug mask */
    
    /* *** Block/CD-ROM device *** */
    int device_users; /* Number of block device users */
    spinlock_t lock; /* Spinlock */
    struct request_queue *queue; /* Device request queue */
    struct gendisk *gd; /* Gendisk structure */
    struct cdrom_device_info cdi; /* CD-ROM driver structue */
    
    struct request *current_request; /* Request to be processed */
    struct mutex req_mutex; /* Request access mutex */
    struct request_sense tmp_sense; /* Temporary sense, in case request doesn't provide one */
    int status; /* Return status of current request */
    wait_queue_head_t request_wait; /* Request wait queue */
    
    /* *** Linux device model *** */
    struct device ldm_device; /* For inclusion on virtual bus */
    struct cdemu_driver *driver; /* Driver */
    
    /* *** Control device *** */
    struct cdev cdev; /* Control character device */
    struct class_device *ctl_class_device; /* Control class device */
    int ctl_opened; /* Is control device opened? */
    int stage; /* Stage of request processing */
    wait_queue_head_t read_wait; /* Read wait queue */
    wait_queue_head_t write_wait; /* Write wait queue */
};
#define to_cdemu_device(dev) container_of(dev, struct cdemu_device, ldm_device);


/* Request processing stages; also lseek offsets */
#define STAGE_CMD_READ_READY    0xFEED0001 /* Command's CDB ready to be read */
#define STAGE_DATA_READ_READY   0xFEED0002 /* Command's data ready to be read */
#define STAGE_DATA_WRITE_WAIT   0xFEED0003 /* Waiting for return data to be written */
#define STAGE_SENSE_WRITE_WAIT  0xFEED0004 /* Waiting for command sense to be written */
#define STAGE_STATUS_WRITE_WAIT 0xFEED0005 /* Waiting for command return status to be written */
#define STAGE_REQUEST_COMPLETE  0xFEED0006 /* Request completed */
#define STAGE_DEVICE_READY      0xFEED0007 /* Device ready to accept new request */
#define STAGE_DEVICE_BUSY       0xFEEDBEEF /* Device busy */

/* Borrowed from ide-cd.h */
#define ATAPI_WAIT_PC         (60 * HZ)
#define ATAPI_WAIT_WRITE_BUSY (10 * HZ)

/* ATAPI sense keys (from table 140 of ATAPI 2.6) */
#define NO_SENSE        0x00
#define RECOVERED_ERROR 0x01
#define NOT_READY       0x02
#define MEDIUM_ERROR    0x03
#define HARDWARE_ERROR  0x04
#define ILLEGAL_REQUEST 0x05
#define UNIT_ATTENTION  0x06
#define DATA_PROTECT    0x07
#define BLANK_CHECK     0x08
#define ABORTED_COMMAND 0x0B
#define MISCOMPARE      0x0E

/* Debug macro and debug level masks */
#define MODULE_DEBUG_ERROR   0xF1 /* Error; maps to KERN_ERR */
#define MODULE_DEBUG_WARNING 0xF2 /* Warning; maps to KERN_WARNING */
/* All masks map to KERN_DEBUG as they're tied to debug messages */
#define MODULE_DEBUG_CTLDEV  0x01 /* Control device functions related */
#define MODULE_DEBUG_BLKDEV  0x02 /* Block device functions related */
#define MODULE_DEBUG_CDROM   0x04 /* CD-ROM device functions related */
#define MODULE_DEBUG_REQUEST 0x08 /* Request queue related */
#define MODULE_DEBUG_AUX     0x10 /* Auxiliary CD-ROM functions related */

#define CDEMU_DEBUG_DEV(dev, lvl, msg...) {     \
    if (lvl == MODULE_DEBUG_ERROR) {            \
        printk(KERN_ERR msg);                   \
    } else if (lvl == MODULE_DEBUG_WARNING) {   \
        printk(KERN_WARNING msg);               \
    } else if (dev->debug_mask & lvl) {         \
        printk(KERN_DEBUG msg);                 \
    }                                           \
}

#define CDEMU_DEBUG_DRV(lvl, msg...) {          \
    if (lvl == MODULE_DEBUG_ERROR) {            \
        printk(KERN_ERR msg);                   \
    } else if (lvl == MODULE_DEBUG_WARNING) {   \
        printk(KERN_WARNING msg);               \
    }                                           \
}

/* Module information */
MODULE_AUTHOR("Rok Mandeljc <rok.mandeljc@email.si");
MODULE_DESCRIPTION("CD Emulator for CD-ROM images ver " CDEMU_MODULE_VERSION);
MODULE_LICENSE("GPL");

/* Module parameters */
static int param_devices = 1;
module_param(param_devices, int, 0);
MODULE_PARM_DESC(param_devices, "Number of virtual CD-ROM devices to create [1-255]");

static int param_ctl_major = 0;
module_param(param_ctl_major, int, 0);
MODULE_PARM_DESC(param_ctl_major, "Major for CDEmu control devices [0-255]");

static int param_block_major = 0;
module_param(param_block_major, int, 0);
MODULE_PARM_DESC(param_ctl_major, "Major for CDEmu block devices [0-255]");


static struct cdemu_device *devices = NULL;


/* ************************************************************************** *\
 *                        CDEMU Linux Device Model Part                       *
\* ************************************************************************** */
/* Bus */
static int cdemu_bus_match (struct device *dev, struct device_driver *driver) {
    /* Simple test; match 'cdemu' in device and driver */
    return !(strncmp(dev->bus_id, "cdemu", strlen("cdemu")) && strncmp(driver->name, "cdemu", strlen("cdemu")));
}

struct bus_type cdemu_bus_type = {
    .name  = "cdemu-bus",
    .match = cdemu_bus_match
};

/* Bus device */
static void cdemu_busdev_release (struct device *dev) {
    return;
}

static struct device cdemu_busdev = {
    .bus_id  = "cdemu-busdev",
    .release = cdemu_busdev_release
};

static int cdemu_bus_create (void) {
    int ret = 0;

    ret = bus_register(&cdemu_bus_type);
    if (ret) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to register-bus!\n", __func__);
        return ret;
    }
    
    ret = device_register(&cdemu_busdev);
    if (ret) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to register bus device!\n", __func__);
        return ret;
    }
    
    return ret;
}

static void cdemu_bus_destroy (void) {
    device_unregister(&cdemu_busdev);
    bus_unregister(&cdemu_bus_type);
}

/* Driver */
static struct cdemu_driver cdemu_driver = {
    .version = CDEMU_MODULE_VERSION,
    .module  = THIS_MODULE,
    .driver  = {
        .name = "cdemu-driver",
    },
};

static ssize_t cdemu_driver_version_show (struct device_driver *driver, char *buf) {
    return sprintf(buf, "%s", CDEMU_MODULE_VERSION) + 1;
}

static ssize_t cdemu_driver_devices_show (struct device_driver *driver, char *buf) {
    struct cdemu_driver *cdemu_driver = to_cdemu_driver(driver);
    return sprintf(buf, "%i", cdemu_driver->devices) + 1;
}

static ssize_t cdemu_driver_ctl_major_show (struct device_driver *driver, char *buf) {
    struct cdemu_driver *cdemu_driver = to_cdemu_driver(driver);
    return sprintf(buf, "%i", cdemu_driver->ctl_major) + 1;
}

static ssize_t cdemu_driver_block_major_show (struct device_driver *driver, char *buf) {
    struct cdemu_driver *cdemu_driver = to_cdemu_driver(driver);
    return sprintf(buf, "%i", cdemu_driver->block_major) + 1;
}

static ssize_t cdemu_driver_debug_masks_show (struct device_driver *driver, char *buf) {
    /*struct cdemu_driver *cdemu_driver = to_cdemu_driver(driver);*/
    int ret = 0;
    int i;
    
    struct {
        char *name;
        int mask;
    } debug_masks[] = {
        { "MODULE_DEBUG_CTLDEV", MODULE_DEBUG_CTLDEV },
        { "MODULE_DEBUG_BLKDEV", MODULE_DEBUG_BLKDEV },
        { "MODULE_DEBUG_CDROM", MODULE_DEBUG_CDROM },
        { "MODULE_DEBUG_REQUEST", MODULE_DEBUG_REQUEST },
        { "MODULE_DEBUG_AUX", MODULE_DEBUG_AUX },
    };
    
    for (i = 0; i < sizeof(debug_masks)/sizeof(debug_masks[0]); i++) {
        ret += sprintf(buf+ret, "%s:0x%02X;\n", debug_masks[i].name, debug_masks[i].mask);
    }
    
    return ret + 1;
}

static DRIVER_ATTR(version, 0444, cdemu_driver_version_show, NULL);
static DRIVER_ATTR(devices, 0444, cdemu_driver_devices_show, NULL);
static DRIVER_ATTR(ctl_major, 0444, cdemu_driver_ctl_major_show, NULL);
static DRIVER_ATTR(block_major, 0444, cdemu_driver_block_major_show, NULL);
static DRIVER_ATTR(debug_masks, 0444, cdemu_driver_debug_masks_show, NULL);

static int cdemu_driver_create (void) {
    int result = 0;
    dev_t devno = MKDEV(param_ctl_major, 0);
    
    /* Register character control device */
    if (param_ctl_major) {
        /* Static major number allocation */
        result = register_chrdev_region(devno, param_devices, "cdemu_ctl");
    } else {
        /* Dynamic major number allocation */
        result = alloc_chrdev_region(&devno, 0, param_devices, "cdemu_ctl");
        param_ctl_major = MAJOR(devno);
    }
    if (result < 0) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to get major for ctl device (%d)!\n", __func__, param_ctl_major);
        return result;
    }
    
    /* Create class for control devices */
    cdemu_driver.cdemu_ctl_class = class_create(THIS_MODULE, "cdemu_ctl");
    
    /* Register block device */
    param_block_major = register_blkdev(param_block_major, "cdemu");
    if (param_block_major <= 0) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to get major for block device!\n", __func__);
        return -EBUSY;
    }
    
    /* Store the parameters */
    cdemu_driver.devices = param_devices;
    cdemu_driver.ctl_major = param_ctl_major;
    cdemu_driver.block_major = param_block_major;
    
    /* Register driver with bus */
    cdemu_driver.driver.bus = &cdemu_bus_type;
    result = driver_register(&cdemu_driver.driver);
    if (result) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to register cdemu driver!\n", __func__);
        return result;
    }
    
    /* Attribute: version */
    result = driver_create_file(&cdemu_driver.driver, &driver_attr_version);
    if (result) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to create 'version' attribute file!\n", __func__);
        return result;
    }
    
    /* Attribute: devices */
    result = driver_create_file(&cdemu_driver.driver, &driver_attr_devices);
    if (result) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to create 'devices' attribute file!\n", __func__);
        return result;
    }
    
    /* Attribute: ctl_major */
    result = driver_create_file(&cdemu_driver.driver, &driver_attr_ctl_major);
    if (result) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to create 'ctl_major' attribute file!\n", __func__);
        return result;
    }
    
    /* Attribute: block_major */
    result = driver_create_file(&cdemu_driver.driver, &driver_attr_block_major);
    if (result) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to create 'block_major' attribute file!\n", __func__);
        return result;
    }
    
    /* Attribute: debug_masks */
    result = driver_create_file(&cdemu_driver.driver, &driver_attr_debug_masks);
    if (result) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to create 'debug_masks' attribute file!\n", __func__);
        return result;
    }
    
    return result;
}

static void cdemu_driver_destroy (void) {
    /* Unregister character control devices */
    dev_t devno = MKDEV(cdemu_driver.ctl_major, 0);
    unregister_chrdev_region(devno, cdemu_driver.devices);
    /* Destroy class for control devices */
    class_destroy(cdemu_driver.cdemu_ctl_class);
    /* Unregister block devices */
    unregister_blkdev(cdemu_driver.block_major, "cdemu");
    /* Remove attribute files */
    driver_remove_file(&cdemu_driver.driver, &driver_attr_version);
    driver_remove_file(&cdemu_driver.driver, &driver_attr_devices);
    driver_remove_file(&cdemu_driver.driver, &driver_attr_ctl_major);
    driver_remove_file(&cdemu_driver.driver, &driver_attr_block_major);
    driver_remove_file(&cdemu_driver.driver, &driver_attr_debug_masks);
    /* Unregister driver */
    driver_unregister(&cdemu_driver.driver);
    
    return;
}

/* Device */
static void cdemu_device_release (struct device *dev) {
    struct cdemu_device *cdemu_device = (struct cdemu_device *)dev_get_drvdata(dev);
	CDEMU_DEBUG_DEV(cdemu_device, MODULE_DEBUG_WARNING, "%s: not implemented yet!\n", __func__);
    /* Release function; basically, we could do all the cleanup here */
    return;
}

static ssize_t cdemu_device_stage_show (struct device *dev, struct device_attribute *attr, char *buf) {
    struct cdemu_device *cdemu_device = to_cdemu_device(dev);
    return sprintf(buf, "0x%X", cdemu_device->stage) + 1;
}

static ssize_t cdemu_device_debug_mask_show (struct device *dev, struct device_attribute *attr, char *buf) {
    struct cdemu_device *cdemu_device = to_cdemu_device(dev);
    return sprintf(buf, "0x%X", cdemu_device->debug_mask) + 1;
}

static ssize_t cdemu_device_debug_mask_store (struct device *dev, struct device_attribute *attr, const char *buf, size_t count) {
    struct cdemu_device *cdemu_device = to_cdemu_device(dev);
    cdemu_device->debug_mask = simple_strtol(buf, NULL, 0);
    return count;
}

static DEVICE_ATTR(stage, 0444, cdemu_device_stage_show, NULL);
static DEVICE_ATTR(debug_mask, 0664, cdemu_device_debug_mask_show,  cdemu_device_debug_mask_store);


/* ************************************************************************** *\
 *                           CDEMU Block Device Part                          *
\* ************************************************************************** */
static int cdemu_block_open (struct inode *inode, struct file *file) {
    struct cdemu_device *dev = inode->i_bdev->bd_disk->private_data;
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_BLKDEV, "%s: start (inode: %p, file: %p)\n", __func__, inode, file);
    file->private_data = dev;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_BLKDEV, "%s: calling cdrom_open\n", __func__);
    ret = cdrom_open(&dev->cdi, inode, file);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_BLKDEV, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_block_release (struct inode *inode, struct file *file) {
    struct cdemu_device *dev = inode->i_bdev->bd_disk->private_data;
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_BLKDEV, "%s: start (inode: %p, file: %p)\n", __func__, inode, file);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_BLKDEV, "%s: calling cdrom_release\n", __func__);
    ret = cdrom_release(&dev->cdi, file);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_BLKDEV, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_block_media_changed (struct gendisk *gd) {
    struct cdemu_device *dev = gd->private_data;
    int ret = 0;

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_BLKDEV, "%s: start (gendisk: %p)\n", __func__, gd);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_BLKDEV, "%s: calling cdrom_media_changed\n", __func__);
    ret = cdrom_media_changed(&dev->cdi);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_BLKDEV, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_block_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg) {
    struct cdemu_device *dev = inode->i_bdev->bd_disk->private_data;
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_BLKDEV, "%s: start (inode: %p, file: %p, cmd: 0x%X, arg: 0x%lX)\n", __func__, inode, file, cmd, arg);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_BLKDEV, "%s: calling cdrom_ioctl\n", __func__);
    ret = cdrom_ioctl(file, &dev->cdi, inode, cmd, arg);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_BLKDEV, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static struct block_device_operations cdemu_block_ops = {
    .owner         = THIS_MODULE,
    .open          = cdemu_block_open,
    .release       = cdemu_block_release,
    .media_changed = cdemu_block_media_changed,
    .ioctl         = cdemu_block_ioctl
};


/* ************************************************************************** *\
 *                       CDEMU Block Request Queue Part                       *
\* ************************************************************************** */
static int cdemu_end_request (struct cdemu_device *dev, struct request *req, int succeeded) {
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: start (dev: %p, req: %p, succeeded: %i)\n", __func__, dev, req, succeeded);

    /* If it's FS request, it succeeded and yet still has BIO, it means it's not
       done yet... */
    if (blk_fs_request(req) && succeeded && req->bio) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: FS request not finished yet, requeing.\n", __func__);
        blk_requeue_request(dev->queue, req);        
        return 0;
    }
    
    /* Sense magic if request failed */
    if (!succeeded) {
        struct request_sense *sense = req->sense;
        /* Well, if sense_len = 0, this means userspace didn't give us sense...
           And since we failed, it might as well mean userspace daemon is not
           attached; so we need to set default error */
        if (req->sense_len == 0 && req->sense) {
            sense->error_code = 0x70;
            sense->valid = 0;
            sense->sense_key = 0x02; /* Sense key */
            sense->add_sense_len = 0x0A; /* Additional sense length */
            /* We return MEDIUM NOT PRESENT, because it keeps TEST UNIT READY 
               happy (even though something along lines of LOGICAL UNIT NOT 
               READY, CAUSE NOT REPORTABLE would be more fitting >.<)*/
            sense->asc = 0x3A;
            sense->ascq = 0x00;
            req->sense_len = 19;
        }
        
        /* Shrink the capacity if requested sector was out of range */
        if (sense && sense->error_code == 0x70) {
            unsigned long sector = 0;
            unsigned long bio_sectors = 0;
            
            switch(sense->sense_key) {
                case MEDIUM_ERROR:
                case ILLEGAL_REQUEST: {
                    if (!sense->valid)
                        break;
                    if (!blk_fs_request(req))
                        break;
                    sector = (sense->information[0] << 24) |
                             (sense->information[1] << 16) |
                             (sense->information[2] <<  8) |
                             (sense->information[3]);
                    
                    bio_sectors = bio_sectors(req->bio);
                    if (bio_sectors < 4)
                        bio_sectors = 4;
                    if (dev->queue->hardsect_size == 2048)
                        sector <<= 2; /* Device sector size is 2K */
                    sector &= ~(bio_sectors -1);
                                        
                    if (sector < get_capacity(dev->gd)) {
                        set_capacity(dev->gd, sector);
                    }
                }
            }
        }
        
        /* Cleanup BIOs... */
        if (blk_fs_request(req)) {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: FS request failed... cleaning up BIOs (%li).\n", __func__, req->nr_sectors);
            end_that_request_first(req, 0, req->nr_sectors);
        }
    }
    
    /* End request - plain and simple */
    end_that_request_last(req, succeeded);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: end ()\n", __func__);
    return 0;
}

static void cdemu_process_pc_request (struct cdemu_device *dev, struct request *req) {    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: start (dev: %p, req: %p)\n", __func__, dev, req);

    /* Prepare status; we fail by default */
    dev->status = 0x02; /* CHECK CONDITION */    
    
    dev->stage = STAGE_CMD_READ_READY;
    /* Wakey wakey */
    wake_up_interruptible(&dev->read_wait);
    wake_up_interruptible(&dev->write_wait);
        
    /* Well, no point in waiting if we don't have userspace daemon running... */
    if (dev->ctl_opened) {
        /* Sleep and pray that stuff gets done */
        if (!wait_event_timeout(dev->request_wait, dev->stage == STAGE_REQUEST_COMPLETE, ATAPI_WAIT_PC)) {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_WARNING, "%s: timed out (while in stage 0x%X)!\n", __func__, dev->stage);
        }
    } else CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: userspace daemon not attached!\n", __func__);
    
    
    /* End the command */
    if (mutex_lock_interruptible(&dev->req_mutex) < 0) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: failed to acquire req access semaphore!\n", __func__);
        /* FIXME: error? */
        return;
    }
    
    if (dev->stage == STAGE_REQUEST_COMPLETE && !dev->status) {
        /* Command succeeded */
        cdemu_end_request(dev, req, 1);
    } else {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: packet command 0x%02X failed!\n", __func__, req->cmd[0]);
        /* Command failed */
        req->errors = dev->status;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
        req->flags |= REQ_FAILED;
#else
        req->cmd_flags |= REQ_FAILED;
#endif
        
        cdemu_end_request(dev, req, 0);
    }
    dev->stage = STAGE_DEVICE_READY;
    dev->current_request = NULL;
    
    mutex_unlock(&dev->req_mutex);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: end ()\n", __func__);

    return;
}

static int cdemu_block_prep_rq (request_queue_t *q, struct request *req) {
    struct cdemu_device *dev = q->queuedata;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: start (q: %p, req: %p)\n", __func__, q, req);

    /* Assign temporary sense if it's not provided... */
    if (!req->sense) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: setting up temporary sense.\n", __func__);
        memset(&dev->tmp_sense, 0, sizeof(struct request_sense));
        req->sense = &dev->tmp_sense;
        req->sense_len = 0;
    }
    
    /* FS request -> READ 10 packet command... */
    if (blk_fs_request(req)) {
        struct bio* bio = req->bio;    
        int hard_sect = 0;    
        long sector = 0;
        unsigned long sectors = 0;
        
        if (!bio) {
            /* Shouldn't really happen... */
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: no BIO in FS request!\n", __func__);
            return BLKPREP_KILL;
        }
        
        /* Convert one BIO at time... */
        hard_sect = queue_hardsect_size(q);
        sector = (long)req->bio->bi_sector / (hard_sect >> 9);
        sectors = bio_sectors(req->bio) / (hard_sect >> 9);

        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: block: %li -> sector: %li, blocks: %i -> sectors: %li\n", __func__, (long)req->bio->bi_sector, sector, bio_sectors(req->bio), sectors);
        
        /* I want to see it if/when it happens */
        if ((long)req->bio->bi_sector % 4) {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_WARNING, "%s: non-aligned sector (0x%lX)!\n", __func__, (long)req->bio->bi_sector);
        }
        if (bio_sectors(req->bio) % 4) {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_WARNING, "%s: non-aligned number of sectors (0x%lX)!\n", __func__, (long)req->bio->bi_sector);
        }
        
        /* Prepare CDB */
        memset(req->cmd, 0, BLK_MAX_CDB);
        /* READ 10 for reading, WRITE 10 for writing */
        if (rq_data_dir(req) == READ) {
            req->cmd[0] = GPCMD_READ_10;
        } else {
            req->cmd[0] = GPCMD_WRITE_10;
        }
        req->cmd[1] = 0;
        /* Sector */
        req->cmd[2] = (sector & 0xFF000000) >> 24;
        req->cmd[3] = (sector & 0x00FF0000) >> 16;
        req->cmd[4] = (sector & 0x0000FF00) >> 8;
        req->cmd[5] = (sector & 0x000000FF) >> 0;
        /* Length */
        req->cmd[7] = (sectors & 0xFF00) >> 8;
        req->cmd[8] = (sectors & 0x00FF) >> 0;
        req->cmd_len = 10;
        
        /*req->flags |= REQ_QUIET;*/
    }
    

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
    if (req->cmd_type != REQ_TYPE_BLOCK_PC && req->cmd_type != REQ_TYPE_FS) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_WARNING, "%s: request of type 0x%X not handled yet!\n", __func__, req->cmd_type);        
    }
#endif

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: end (BLKPREP_OK)\n", __func__);
    return BLKPREP_OK;
}

static void cdemu_block_request (request_queue_t *q) {
    struct cdemu_device *dev = q->queuedata;
    struct request *req;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: start (q: %p)\n", __func__, q);

    /* I think this can happen because of preemption... */
    if (dev->stage != STAGE_DEVICE_READY) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: gimme a break, can't you see I'm busy?!\n", __func__);
        return;
    }
    
    /* Endless loop, broken only by lack of requests... */
    while (1) {
        spin_unlock_irq(&dev->lock);

        /* We're busy now */        
        dev->stage = STAGE_DEVICE_BUSY;

        /* Fetch next request */        
        req = elv_next_request(dev->queue);
        /* No request */
        if (!req) {
            dev->stage = STAGE_DEVICE_READY;
            spin_lock_irq(&dev->lock);
            break;
        }
        
        if (mutex_lock_interruptible(&dev->req_mutex)) {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: failed to acquire req access semaphore!\n", __func__);
            /* FIXME: error? */
            spin_lock_irq(&dev->lock);			
            return;
        }
        /* Set current request */
        blkdev_dequeue_request(req);
        dev->current_request = req;
        mutex_unlock(&dev->req_mutex);
        
        /* *** Process the request *** */
        cdemu_process_pc_request(dev, req);
        
        /* Lock back */
        spin_lock_irq(&dev->lock);
    }
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_REQUEST, "%s: end ()\n", __func__);
    return;
}


/* ************************************************************************** *\
 *               CDEMU Auxiliary functions for CD-ROM Device Part             *
 *                  (implemented with generic packet interface)               *
\* ************************************************************************** */
static int cdemu_aux_check_status (struct cdrom_device_info *cdi, struct request_sense *sense) {
    struct cdemu_device *dev = cdi->handle;
    struct packet_command cgc;
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: start (cdi: %p, sense: %p)\n", __func__, cdi, sense);
    
    init_cdrom_command(&cgc, NULL, 0, CGC_DATA_NONE);
    cgc.sense = sense;
    
    /* Prepare CDB */
    cgc.cmd[0] = GPCMD_TEST_UNIT_READY;

    /* Send it as generic packet */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: queuing command...\n", __func__);
    ret = cdi->ops->generic_packet(cdi, &cgc);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_aux_lockdoor (struct cdrom_device_info *cdi, int lockflag, struct request_sense *sense) {
    struct cdemu_device *dev = cdi->handle;
    struct packet_command cgc;
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: start (cdi: %p, lockflag: %i, sense: %p)\n", __func__, cdi, lockflag, sense);
    
    init_cdrom_command(&cgc, NULL, 0, CGC_DATA_NONE);
    cgc.sense = sense;
    
    /* Prepare CDB */
    cgc.cmd[0] = GPCMD_PREVENT_ALLOW_MEDIUM_REMOVAL;
    cgc.cmd[4] = lockflag;

    /* Send it as generic packet */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: queuing command...\n", __func__);
    ret = cdi->ops->generic_packet(cdi, &cgc);
    
    /* No medium, that's alright. */
    if (ret != 0 && sense->sense_key == NOT_READY && sense->asc == 0x3a) {
        ret = 0;
    }

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_aux_eject (struct cdrom_device_info *cdi, int ejectflag, struct request_sense *sense) {
    struct cdemu_device *dev = cdi->handle;
    struct packet_command cgc;
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: start (cdi: %p, ejectflag: %i, sense: %p)\n", __func__, cdi, ejectflag, sense);
    
    init_cdrom_command(&cgc, NULL, 0, CGC_DATA_NONE);
    cgc.sense = sense;
    
    /* Prepare CDB */
    cgc.cmd[0] = GPCMD_START_STOP_UNIT;
    cgc.cmd[1] = 2 | ejectflag;
    
    /* Send it as generic packet */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: queuing command...\n", __func__);
    ret = cdi->ops->generic_packet(cdi, &cgc);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_aux_select_speed (struct cdrom_device_info *cdi, int speed, struct request_sense *sense) {
    struct cdemu_device *dev = cdi->handle;
    struct packet_command cgc;
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: start (cdi: %p, speed: %i, sense: %p)\n", __func__, cdi, speed, sense);

    if (speed == 0) {
        speed = 0xFFFF; /* Set to max */
    } else {
        speed *= 177;   /* Nx to kbytes/s */
    }
    
    init_cdrom_command(&cgc, NULL, 0, CGC_DATA_NONE);
    cgc.sense = sense;
    
    /* Prepare CDB */
    cgc.cmd[0] = GPCMD_SET_SPEED;
    cgc.cmd[2] = (speed & 0xFF00) >> 8;    
    cgc.cmd[3] = speed & 0xFF;
    cgc.cmd[4] = (speed & 0xFF00) >> 8;    
    cgc.cmd[5] = speed & 0xFF;

    /* Send it as generic packet */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: queuing command...\n", __func__);
    ret = cdi->ops->generic_packet(cdi, &cgc);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_aux_multisession_info (struct cdrom_device_info *cdi, unsigned *lba, unsigned char *xa, struct request_sense *sense) {
    struct cdemu_device *dev = cdi->handle;
    struct packet_command cgc;
    unsigned char buffer[12] = "";
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: start (cdi: %p, lba: %p, xa: %p, sense: %p)\n", __func__, cdi, lba, xa, sense);

    init_cdrom_command(&cgc, buffer, sizeof(buffer), CGC_DATA_READ);
    cgc.sense = sense;
    
    /* Prepare CDB */
    cgc.cmd[0] = GPCMD_READ_TOC_PMA_ATIP;
    cgc.cmd[2] = 2;
    cgc.cmd[7] = (sizeof(buffer) & 0xFFFF0000) >> 8;
    cgc.cmd[8] = (sizeof(buffer) & 0x0000FFFF) >> 0;

    /* Send it as generic packet */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: queuing command...\n", __func__);
    ret = cdi->ops->generic_packet(cdi, &cgc);
    
    if (!ret) {
        if (lba) {
            *lba = (buffer[8] << 24) | buffer[9] << 16 | buffer[10] << 8 | buffer[11];
        }
        if (xa) {
            *xa = (buffer[2] != buffer[3]); /* If first and last session are different */
        }
    }
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_aux_read_subchannel (struct cdrom_device_info *cdi, int format, char *buf, int buflen, struct request_sense *sense) {
    struct cdemu_device *dev = cdi->handle;
    struct packet_command cgc;
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: start (cdi: %p, format: %i, buf: %p, buflen: %i, sense: %p)\n", __func__, cdi, format, buf, buflen, sense);

    init_cdrom_command(&cgc, buf, buflen, CGC_DATA_READ);
    cgc.sense = sense;
    
    /* Prepare CDB */
    cgc.cmd[0] = GPCMD_READ_SUBCHANNEL;
    cgc.cmd[1] = 2; /* MSF */
    cgc.cmd[2] = 64; /* Request subQ data */
    cgc.cmd[3] = format;
    cgc.cmd[7] = (buflen & 0xFFFF0000) >> 8;
    cgc.cmd[8] = buflen & 0x0000FFFF;

    /* Send it as generic packet */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: queuing command...\n", __func__);
    ret = cdi->ops->generic_packet(cdi, &cgc);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_aux_toc_header (struct cdrom_device_info *cdi, struct cdrom_tochdr *tochdr, struct request_sense *sense) {
    struct cdemu_device *dev = cdi->handle;
    struct packet_command cgc;
    unsigned char buffer[4] = "";
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: start (cdi: %p, tochdr: %p, sense: %p)\n", __func__, cdi, tochdr, sense);

    init_cdrom_command(&cgc, buffer, sizeof(buffer), CGC_DATA_READ);
    cgc.sense = sense;
    
    /* Prepare CDB */
    cgc.cmd[0] = GPCMD_READ_TOC_PMA_ATIP;
    cgc.cmd[2] = 0; /* Format: 0 */
    cgc.cmd[7] = (sizeof(buffer) & 0xFFFF000) >> 8;
    cgc.cmd[8] = sizeof(buffer) & 0x0000FFFF;

    /* Send it as generic packet */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: queuing command...\n", __func__);
    ret = cdi->ops->generic_packet(cdi, &cgc);
    
    if (!ret) {
        tochdr->cdth_trk0 = buffer[2];
        tochdr->cdth_trk1 = buffer[3];
    }

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_aux_toc_entry (struct cdrom_device_info *cdi, struct cdrom_tocentry *tocentry, struct request_sense *sense) {
    struct cdemu_device *dev = cdi->handle;
    struct packet_command cgc;
    unsigned char buffer[12] = ""; /* Header + one descriptor */
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: start (cdi: %p, tocentry: %p, sense: %p)\n", __func__, cdi, tocentry, sense);

    init_cdrom_command(&cgc, buffer, sizeof(buffer), CGC_DATA_READ);
    cgc.sense = sense;
    
    /* Prepare CDB */
    cgc.cmd[0] = GPCMD_READ_TOC_PMA_ATIP;
    cgc.cmd[1] = (tocentry->cdte_format == CDROM_MSF) << 1;
    cgc.cmd[2] = 0;
    cgc.cmd[6] = tocentry->cdte_track;
    cgc.cmd[7] = (sizeof(buffer) & 0xFFFF0000) >> 8;
    cgc.cmd[8] = sizeof(buffer) & 0x0000FFFF;

    /* Send it as generic packet */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: queuing command...\n", __func__);
    ret = cdi->ops->generic_packet(cdi, &cgc);
    
    if (!ret) {
        tocentry->cdte_adr = (buffer[5] & 0xF0) >> 4;
        tocentry->cdte_ctrl = buffer[5] & 0x0F;
        if (tocentry->cdte_format == CDROM_MSF) {
            tocentry->cdte_addr.msf.minute = buffer[9];
            tocentry->cdte_addr.msf.second = buffer[10];
            tocentry->cdte_addr.msf.frame = buffer[11];
        } else {
            tocentry->cdte_addr.lba = (buffer[8] << 24) | (buffer[9] << 16) | (buffer[10] << 8) | buffer[11];
        }
    }

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}


static int cdemu_aux_read_capacity (struct cdrom_device_info *cdi, unsigned long *capacity, unsigned long *spf, struct request_sense *sense) {
    struct cdemu_device *dev = cdi->handle;
    struct packet_command cgc;
    struct { __u32 lba; __u32 blocklen; } buffer = {0, 0};
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: start (cdi: %p, capacity: %p, spf: %p, sense: %p)\n", __func__, cdi, capacity, spf, sense);

    init_cdrom_command(&cgc, &buffer, sizeof(buffer), CGC_DATA_READ);
    cgc.sense = sense;
    
    /* Prepare CDB */
    cgc.cmd[0] = GPCMD_READ_CDVD_CAPACITY;
    
    /* Send it as generic packet */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: queuing command...\n", __func__);
    ret = cdi->ops->generic_packet(cdi, &cgc);
    
    if (!ret) {
        if (capacity) {
            *capacity = 1 + be32_to_cpu(buffer.lba);
        }
        if (spf) {
            *spf = be32_to_cpu(buffer.blocklen) >> 9;
        }
    }
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_AUX, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

/* ************************************************************************** *\
 *                          CDEMU CD-ROM Device Part                          *
\* ************************************************************************** */
static int cdemu_cdrom_open (struct cdrom_device_info *cdi, int purpose) {
    struct cdemu_device *dev = cdi->handle;
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: start (cdi: %p, purpose: %i)\n", __func__, cdi, purpose);
    
    dev->device_users++;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: end (ret: %i, dev->device_users: %i)\n", __func__, ret, dev->device_users);
    return ret;
}

static void cdemu_cdrom_release (struct cdrom_device_info *cdi) {
    struct cdemu_device *dev = cdi->handle;
    int ret = 0;

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: start (cdi: %p)\n", __func__, cdi);
    
    dev->device_users--;

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: end (ret: %i, dev->device_users: %i)\n", __func__, ret, dev->device_users);
    return;    
}

static int cdemu_cdrom_drive_status (struct cdrom_device_info *cdi, int slot_nr) {
    struct cdemu_device *dev = cdi->handle;
    struct media_event_desc med;
    struct request_sense sense;
    int ret = 0;

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: start (cdi: %p, slot_nr: %i)\n", __func__, cdi, slot_nr);

    if (slot_nr != CDSL_CURRENT) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: slot_nr not CDSL_CURRENT!\n", __func__);
        ret = -EINVAL;
        goto end;
    }
    
    /* Use check status */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: calling cdemu_aux_check_status\n", __func__);
    ret = cdemu_aux_check_status(cdi, &sense);
    if (!ret || sense.sense_key == UNIT_ATTENTION) {
        ret = CDS_DISC_OK;
        goto end;
    }
    
    /* Use get media event */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: calling cdrom_get_media_event\n", __func__);
    if (!cdrom_get_media_event(cdi, &med)) {
        if (med.media_present) {
            ret = CDS_DISC_OK;
            goto end;
        } else if (med.door_open) {
            ret = CDS_TRAY_OPEN;
            goto end;
        } else {
            ret = CDS_NO_DISC;
            goto end;
        }
    }

    if (sense.sense_key == NOT_READY && sense.asc == 0x04 && sense.ascq == 0x04) {
        ret = CDS_DISC_OK;
        goto end;
    }
    
    if (sense.sense_key == NOT_READY) {
        if (sense.asc == 0x3a) {
            if (sense.ascq == 1) {
                ret = CDS_NO_DISC;
                goto end;
            } else if (sense.ascq == 0 || sense.ascq == 2) {
                ret = CDS_TRAY_OPEN;
                goto end;
            }
        }
    }

end:
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_cdrom_media_changed (struct cdrom_device_info *cdi, int slot_nr) {
    struct cdemu_device *dev = cdi->handle;
    struct media_event_desc med;
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: start (cdi: %p, slot_nr: %i)\n", __func__, cdi, slot_nr);
    
    /* Use get media event */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: calling cdrom_get_media_event\n", __func__);
    if (!cdrom_get_media_event(cdi, &med)) {
        if (med.media_event_code == 4) {
            /* Media changed; alright, since we know it changed, maybe we can
               change the capacity here as well? :) */
            unsigned long capacity = 0;
            unsigned long spf = 4; /* Default; 2048/512 */
            struct request_sense sense;
            
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: media changed; reading capacity\n", __func__);
            ret = cdemu_aux_read_capacity(cdi, &capacity, &spf, &sense);
            if (ret != 0) {
                capacity = 0x1FFFFF;
            }
            set_capacity(cdi->disk, capacity * spf);
            
            ret = 1;
        }
    }
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_cdrom_tray_move (struct cdrom_device_info *cdi, int position) {
    struct cdemu_device *dev = cdi->handle;
    struct request_sense sense;
    int ret = 0;

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: start (cdi: %p, position: %i)\n", __func__, cdi, position);

    /* If we're to open tray, unlock the door first */
    if (position) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: calling cdemu_aux_lockdoor\n", __func__);
        ret = cdemu_aux_lockdoor(cdi, 0, &sense);
    }

    /* If all went fine up to this point, do the ejecting/loading */
    if (!ret) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: calling cdemu_aux_eject\n", __func__);
        ret = cdemu_aux_eject(cdi, !position, &sense);
    }
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_cdrom_lock_door (struct cdrom_device_info *cdi, int lock) {
    struct cdemu_device *dev = cdi->handle;
    struct request_sense sense;
    int ret = 0;

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: start (cdi: %p, lock: %i)\n", __func__, cdi, lock);

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: calling cdemu_aux_lockdoor\n", __func__);
    ret = cdemu_aux_lockdoor(cdi, lock, &sense);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_cdrom_select_speed (struct cdrom_device_info *cdi, int speed) {
    struct cdemu_device *dev = cdi->handle;
    struct request_sense sense;
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: start (cdi: %p, speed: %i)\n", __func__, cdi, speed);

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: calling cdemu_aux_select_speed\n", __func__);
    ret = cdemu_aux_select_speed(cdi, speed, &sense);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_cdrom_get_last_session (struct cdrom_device_info *cdi, struct cdrom_multisession *ms_info) {
    struct cdemu_device *dev = cdi->handle;
    struct request_sense sense;
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: start (cdi: %p, ms_info: %p)\n", __func__, cdi, ms_info);

    ms_info->addr_format = CDROM_LBA; /* cdrom_multisession_info returns LBA */
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: calling cdemu_aux_multisession_info\n", __func__);
    ret = cdemu_aux_multisession_info(cdi, &ms_info->addr.lba, &ms_info->xa_flag, &sense);

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: end (ret: %i)\n", __func__, ret);
    return ret;    
}

static int cdemu_cdrom_get_mcn (struct cdrom_device_info *cdi, struct cdrom_mcn *mcn_info) {
    struct cdemu_device *dev = cdi->handle;
    struct request_sense sense;
    char mcnbuf[24] = {0};
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: start (cdi: %p, mcn_info: %p)\n", __func__, cdi, mcn_info);
    
    /* Get MCN */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: calling cdemu_aux_read_subchannel\n", __func__);
    ret = cdemu_aux_read_subchannel(cdi, 2, mcnbuf, sizeof(mcnbuf), &sense);
    
    if (!ret) {
        memcpy(mcn_info->medium_catalog_number, mcnbuf+9, sizeof(mcn_info->medium_catalog_number)-1);
    }

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: end (ret: %i)\n", __func__, ret);
    return ret;    
}

static int cdemu_cdrom_reset (struct cdrom_device_info *cdi) {
    struct cdemu_device *dev = cdi->handle;
    int ret = 0;
    
    /* Nothing to do here */
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: nothing to do here (ret: %i)\n", __func__, ret);

    return ret;
}

static int cdemu_cdrom_audio_ioctl (struct cdrom_device_info *cdi, unsigned int cmd, void *arg) {
    struct cdemu_device *dev = cdi->handle;
    struct request_sense sense;
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: start (cdi: %p, cmd: 0x%X, arg: %p)\n", __func__, cdi, cmd, arg);

    switch (cmd) {
        case CDROMREADTOCHDR: {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: CDROMREADTOCHDR, calling cdemu_aux_toc_header\n", __func__);
            ret = cdemu_aux_toc_header(cdi, (struct cdrom_tochdr *) arg, &sense);
            break;
        }
        case CDROMREADTOCENTRY: {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: CDROMREADTOCENTRY, calling cdemu_aux_toc_entry\n", __func__);
            ret = cdemu_aux_toc_entry(cdi, (struct cdrom_tocentry *) arg, &sense);
            break;
        }
        case CDROMPLAYTRKIND: {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: CDROMPLAYTRKIND, not implemented yet\n", __func__);
            ret = 0;
            break;
        }
        default: {
            printk(KERN_WARNING "%s: Audio IOCTL %X not implemented yet!\n", __func__, cmd);
            ret = -EINVAL;
            break;
        }
    }
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_cdrom_generic_packet (struct cdrom_device_info *cdi, struct packet_command *cgc) {
    struct cdemu_device *dev = (struct cdemu_device *) cdi->handle;
    struct request *req = NULL;
    int ret = 0;

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: start (cdi: %p, cgc: %p, cgc->cmd[0]: 0x%hhX)\n", __func__, cdi, cgc, cgc->cmd[0]);

    /* Allocate request */
    req = blk_get_request(dev->queue, cgc->data_direction == CGC_DATA_WRITE ? WRITE : READ, GFP_KERNEL);
    if (!req) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: failed to create request!\n", __func__);
        return -ENOMEM;
    }
    
    /* Fill in command */
    req->cmd_len = CDROM_PACKET_SIZE;
    memcpy(req->cmd, cgc->cmd, CDROM_PACKET_SIZE);

    /* Sense */
    if (cgc->sense)  {
        memset(cgc->sense, 0, sizeof(struct request_sense));
        req->sense = cgc->sense;
        req->sense_len = 0;
    }
    
    /* Flags */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
    req->flags |= REQ_PC;
    
    if (cgc->quiet) {
        req->flags |= REQ_QUIET;
    }
#else
    req->cmd_type = REQ_TYPE_BLOCK_PC;
    
    if (cgc->quiet) {
        req->cmd_flags |= REQ_QUIET;
    }
#endif
    
    /* BIO buffer map thing */
    if (cgc->data_direction == CGC_DATA_READ || cgc->data_direction == CGC_DATA_WRITE) {
        if (blk_rq_map_kern(dev->queue, req, cgc->buffer, cgc->buflen, __GFP_WAIT)) {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: failed to map BIO!\n", __func__);
            goto out;
        }
    }
    /* No bouncing, since it's kernel buffer and not user one */
    
    /* Timeout */
    req->timeout = cgc->timeout;
    if (req->timeout <= 0) {
        req->timeout = ATAPI_WAIT_PC;
    }
    
    /* Queue packet command */    
    blk_execute_rq(dev->queue, dev->gd, req, 0);
        
    /* Return error if the command failed */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
    if (req->flags & REQ_FAILED) {
#else
    if (req->cmd_flags & REQ_FAILED) {
#endif
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: command failed!\n", __func__);
        ret = -EIO;
    }
    
    cgc->buflen = req->data_len; /* Update length... */
    
    /* There's no need to unmap kernel buffer mappings, right? */
    
out:
    blk_put_request(req);
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CDROM, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

struct cdrom_device_ops cdemu_cdrom_dops = {
    .open               = cdemu_cdrom_open,
    .release            = cdemu_cdrom_release,
    .drive_status       = cdemu_cdrom_drive_status,
    .media_changed      = cdemu_cdrom_media_changed,
    .tray_move          = cdemu_cdrom_tray_move,
    .lock_door          = cdemu_cdrom_lock_door,
    .select_speed       = cdemu_cdrom_select_speed,
    .get_last_session   = cdemu_cdrom_get_last_session,
    .get_mcn            = cdemu_cdrom_get_mcn,
    .reset              = cdemu_cdrom_reset,
    .audio_ioctl        = cdemu_cdrom_audio_ioctl,
    .capability         =    CDC_CLOSE_TRAY | CDC_OPEN_TRAY | CDC_LOCK | 
                             CDC_SELECT_SPEED | CDC_MULTI_SESSION | CDC_MCN | 
                             CDC_MEDIA_CHANGED | CDC_PLAY_AUDIO | CDC_RESET | 
                             CDC_DRIVE_STATUS | CDC_GENERIC_PACKET | CDC_DVD,
    .n_minors           = 1,
    .generic_packet     = cdemu_cdrom_generic_packet,
};


/* ************************************************************************** *\
 *                          CDEMU Control Device Part                         *
\* ************************************************************************** */
static int cdemu_ctl_open(struct inode *inode, struct file *file) {
    struct cdemu_device *dev = container_of(inode->i_cdev, struct cdemu_device, cdev);
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: start (inode: %p, file: %p)\n", __func__, inode, file);

    /* Only one open at time is possible */
    if (dev->ctl_opened) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: device already opened!\n", __func__);
        ret = -EFAULT;
    } else {
        /* And use filp->private_data to point to the device data */
        file->private_data = dev;
        dev->ctl_opened = 1;
    }
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: end (ret: %i)\n", __func__, ret);
    return ret;
}

static int cdemu_ctl_release (struct inode *inode, struct file *file) {
    struct cdemu_device *dev = container_of(inode->i_cdev, struct cdemu_device, cdev);
    int ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: start (inode: %p, file: %p)\n", __func__, inode, file);

    if (dev->ctl_opened) {
        /* Cleanup private data */
        file->private_data = NULL;
        /* Release device */
        dev->ctl_opened = 0;
    }
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: end (ret: %i)\n", __func__, ret);

    return ret;
}

static loff_t cdemu_ctl_lseek (struct file *file, loff_t offset, int origin) {
    struct cdemu_device *dev = file->private_data;
    unsigned long loffset = offset;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: start (file: %p, offset: 0x%llX, origin: %i)\n", __func__, file, offset, origin);
    
    /*  We accept only absolute origin... And offset should have stage prefix...
        Let's be fussy so we have strict interface... Or something :P */
    if (!origin && (loffset & 0xFEED0000)) {
        if (mutex_lock_interruptible(&dev->req_mutex) < 0) {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: failed to acquire req access semaphore!\n", __func__);
            /* FIXME: error? */
            return -EINVAL;
        }
        dev->stage = loffset;
        wake_up_interruptible(&dev->read_wait);
        wake_up_interruptible(&dev->write_wait);
        
        mutex_unlock(&dev->req_mutex);
    } else {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_WARNING, "%s: invalid offset!\n", __func__);
        return -EINVAL;
    }
        
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: end (ret: 0x%X)\n", __func__, dev->stage);
    return dev->stage;
}

static unsigned int cdemu_ctl_poll (struct file *file, poll_table *wait) {
    struct cdemu_device *dev = file->private_data;
    unsigned int mask = 0;

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: start (file: %p, wait: %p)\n", __func__, file, wait);

    /* Wait */
    poll_wait(file, &dev->read_wait, wait);
    poll_wait(file, &dev->write_wait, wait);
    
    /* Check what we got */
    if (dev->stage == STAGE_CMD_READ_READY) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: ready to be read from\n", __func__);
        mask |= POLLIN | POLLRDNORM;
    }
    if (dev->stage == STAGE_DATA_WRITE_WAIT) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: ready to be written to\n", __func__);
        mask |= POLLOUT | POLLWRNORM;
    }
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: end (ret: 0x%X)\n", __func__, mask);
    return mask;
}

static ssize_t cdemu_ctl_read (struct file *file, char __user * buffer, size_t count, loff_t *ppos) {
    struct cdemu_device *dev = file->private_data;
    struct request *req;
    ssize_t transfer_len = 0;
    ssize_t ret = 0;

    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: start (file: %p, buffer: %p, count: 0x%X, ppos: %p)\n", __func__, file, buffer, count, ppos);

    if (count == 0) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: read of 0 bytes requested; returning\n", __func__);
        return 0;
    }
    
    /* Acquire semaphore */
    if (mutex_lock_interruptible(&dev->req_mutex) < 0) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: failed to acquire req access semaphore!\n", __func__);
        /* FIXME: error? */
        return -EFAULT;
    }
    req = dev->current_request;
    if (!req) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: req = NULL!\n", __func__);
        mutex_unlock(&dev->req_mutex);
        return -EFAULT;
    }
    
    switch (dev->stage) {
        case STAGE_CMD_READ_READY: {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: read CDB: %02Xh\n", __func__, req->cmd[0]);
                        
            /* Simple transfer of BLK_MAX_CDB bytes */
            transfer_len = min_t(int, count, BLK_MAX_CDB);
                
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: about to copy %i bytes to userspace\n", __func__, transfer_len);
            if (copy_to_user(buffer, req->cmd, transfer_len)) {
                CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: failed to copy CDB to userspace!\n", __func__);
                mutex_unlock(&dev->req_mutex);
                return -EFAULT;
            }
                
            /* If data direction is write, then we're ready to transfer data
               to device... otherwise it's the other way around */
            if (rq_data_dir(req)) {
                dev->stage = STAGE_DATA_READ_READY;
            } else {
                dev->stage = STAGE_DATA_WRITE_WAIT;
            }
                
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
            dev->current_request->flags |= REQ_STARTED; /* Mark request as started now... */
#else
            dev->current_request->cmd_flags |= REQ_STARTED; /* Mark request as started now... */
#endif
            
            ret = transfer_len;
            
            break;
        }
        case STAGE_DATA_READ_READY: {
            char __user *buf_ptr = buffer;
            size_t thislen = count;

            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: read data\n", __func__);
            
            /* Sanity check; think it shouldn't happen, though... */
            if (!req->bio) {
                CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: data read on request with no BIO!\n", __func__);
                mutex_unlock(&dev->req_mutex);
                return -EINVAL;
            }
            
            /* Do transfer... */
            while (thislen > 0) {
                int req_len = 0;
                char *req_ptr = NULL;
                
                if (!req->bio) {
                    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: no BIO; this means we didn't break when we should have!\n", __func__);
                    mutex_unlock(&dev->req_mutex);
                    return -EINVAL;
                }
                
                req_ptr = bio_data(req->bio);
                req_len = bio_iovec(req->bio)->bv_len;
                
                /* In case there's no buffer to copy to, but we're expected to do it... */
                if (!req_ptr) {
                    if (req_len) {
                        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: no data buffer provided in request even though data length provided\n", __func__);
                        mutex_unlock(&dev->req_mutex);
                        return -EINVAL;
                    }
                }
                
                /* Compare length of remaining data from userspace and size of
                   current BIO's buffer... choose lesser; and then, we do transfers
                   in 1024-byte chunks */
                req_len = min_t(int, req_len, thislen);
                req_len = min_t(int, req_len, 1024);
                
                /* Copy to userspace */
                CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: about to copy %i bytes to userspace\n", __func__, transfer_len);
                req_len -= copy_to_user(buf_ptr, req_ptr, req_len);
                if (!req_len) {
                    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: failed to copy data to userspace!\n", __func__);
                    mutex_unlock(&dev->req_mutex);
                    return -EFAULT;
                }
                
                CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: transfer done\n", __func__);

                thislen -= req_len; /* Update length we use as counter */
                buf_ptr += req_len; /* Adjust buffer pointer */
                transfer_len += req_len; /* Add up transferred length */
                
                /* Update request (cleanup BIOs and all that junk) */
                if (!end_that_request_chunk(req, 1, req_len)) {
                    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: we finished this request (thislen = %i)\n", __func__, thislen);
                    break;
                }
                
            }
                        
            dev->stage = STAGE_SENSE_WRITE_WAIT;
            ret = transfer_len;
            break;
        }
        default: {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_WARNING, "%s: read called at inappropriate stage (0x%X)!\n", __func__, dev->stage);
            mutex_unlock(&dev->req_mutex);
            return -EFAULT;
        }
    }
        
    /* If we gave the user some bytes, update the access time. */
    if (ret > 0) {
        struct inode *inode = file->f_dentry->d_inode;
        inode->i_mtime = current_fs_time(inode->i_sb);
        mark_inode_dirty(inode);

        file_accessed(file);
    }
        
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: end (ret: 0x%X)\n", __func__, ret);
    mutex_unlock(&dev->req_mutex);
    
    return ret;
}

static ssize_t cdemu_ctl_write (struct file *file, const char __user *buffer, size_t count, loff_t *ppos) {
    struct cdemu_device *dev = file->private_data;
    struct request *req;
    ssize_t transfer_len = 0;
    ssize_t ret = 0;
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: start (file: %p, buffer: %p, count: 0x%X, ppos: %p)\n", __func__, file, buffer, count, ppos);
    
    /* Acquire semaphore */
    if (mutex_lock_interruptible(&dev->req_mutex) < 0) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: failed to acquire req access semaphore!\n", __func__);
        /* FIXME: error? */
        return -EFAULT;
    }
    req = dev->current_request;
    if (!req) {
        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: req = NULL!\n", __func__);
        mutex_unlock(&dev->req_mutex);
        return -EFAULT;
    }
    
    switch (dev->stage) {
        case STAGE_DATA_WRITE_WAIT: {
            const char __user *buf_ptr = buffer;
            size_t thislen = count;
            
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: write data\n", __func__);

            /* Sanity check; think it shouldn't happen, though... */
            if (!req->bio) {
                CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: data write on request with no BIO!\n", __func__);
                mutex_unlock(&dev->req_mutex);
                return -EINVAL;
            }
            
            /* Do transfer... */
            while (thislen > 0) {
                int req_len = 0;
                char *req_ptr = NULL;
                
                if (!req->bio) {
                    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: no BIO; this means we didn't break when we should have!\n", __func__);
                    mutex_unlock(&dev->req_mutex);
                    return -EINVAL;
                }
                
                req_ptr = bio_data(req->bio);
                req_len = bio_iovec(req->bio)->bv_len;
                
                /* In case there's no buffer to copy from, but we're expected to do it... */
                if (!req_ptr) {
                    if (req_len) {
                        CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: no data buffer provided in request even though data length provided\n", __func__);
                        mutex_unlock(&dev->req_mutex);
                        return -EINVAL;
                    }
                }
                
                /* Compare length of remaining data from userspace and size of
                   current BIO's buffer... choose lesser; and then, we do transfers
                   in 512-byte chunks */
                req_len = min_t(int, req_len, thislen);
                req_len = min_t(int, req_len, 512);
                CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: about to transfer %i bytes from userspace\n", __func__, req_len);

                /* Copy from userspace */
                req_len -= copy_from_user(req_ptr, buf_ptr, req_len);
                if (!req_len) {
                    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: failed to copy data from userspace!\n", __func__);
                    mutex_unlock(&dev->req_mutex);
                    return -EFAULT;
                }
                
                CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: transfer done\n", __func__);
                                
                thislen -= req_len; /* Update length we use as counter */
                req->data_len -= req_len;
                buf_ptr += req_len; /* Adjust buffer pointer */
                transfer_len += req_len; /* Add up transferred length */
                
                /* Update request (cleanup BIOs and all that junk) */
                if (!end_that_request_chunk(req, 1, req_len)) {
                    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: we finished this request (thislen = %i)\n", __func__, thislen);
                    break;
                }
            }
                        
            dev->stage = STAGE_STATUS_WRITE_WAIT;
            ret = transfer_len;
            break;
        };
        case STAGE_SENSE_WRITE_WAIT: {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: write sense\n", __func__);
            
            transfer_len = min_t(int, count, sizeof(struct request_sense));
            
            if (req->sense) {
                CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: about to copy %i bytes from userspace\n", __func__, transfer_len);
                if (copy_from_user(req->sense, buffer, transfer_len)) {
                    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: failed to copy sense from userspace!\n", __func__);
                    mutex_unlock(&dev->req_mutex);
                    return -EFAULT;
                }
                req->sense_len = transfer_len;
            } else {
                CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_WARNING, "%s: no sense provided (req->sense: %p, req->sense_len: %i)!\n", __func__, req->sense, req->sense_len);
            }
            
            dev->stage = STAGE_STATUS_WRITE_WAIT;
            ret = transfer_len;
            break;
        }
        case STAGE_STATUS_WRITE_WAIT: {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: write status\n", __func__);
            
            transfer_len = min_t(int, count, sizeof(dev->status));
            
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: about to copy %i bytes from userspace\n", __func__, transfer_len);
            if (copy_from_user(&dev->status, buffer, transfer_len)) {
                CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_ERROR, "%s: failed to copy status from userspace!\n", __func__);
                mutex_unlock(&dev->req_mutex);
                return -EFAULT;
            }
            
            dev->stage = STAGE_REQUEST_COMPLETE;
            ret = transfer_len;
            
            /* Wake up, we're done! */
            wake_up(&dev->request_wait);
            break;
        };
        default: {
            CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_WARNING, "%s: write called at inappropriate stage (0x%X)!\n", __func__, dev->stage);
            mutex_unlock(&dev->req_mutex);
            return -EFAULT;
        }
    }
    
    /* If some bytes were written, mark the inode */
    if (ret > 0) {
        struct inode *inode = file->f_dentry->d_inode;
        inode->i_mtime = current_fs_time(inode->i_sb);
        mark_inode_dirty(inode);
        
        file_accessed(file);
    }
    
    CDEMU_DEBUG_DEV(dev, MODULE_DEBUG_CTLDEV, "%s: end (ret: 0x%X)\n", __func__, ret);
    mutex_unlock(&dev->req_mutex);
    
    return ret;
}

struct file_operations cdemu_ctl_fops = {
    .owner   = THIS_MODULE,
    .open    = cdemu_ctl_open,
    .release = cdemu_ctl_release,
    .read    = cdemu_ctl_read,
    .write   = cdemu_ctl_write,
    .llseek  = cdemu_ctl_lseek,
    .poll    = cdemu_ctl_poll,
};


/* ************************************************************************** *\
 *                              CDEMU Module Part                             *
\* ************************************************************************** */
static int cdemu_setup_device (struct cdemu_device *dev, int which) {
    int result;
    int devno = MKDEV(param_ctl_major, which);
    
    memset(dev, 0, sizeof(struct cdemu_device));
    
    /* Control device init */
    cdev_init(&dev->cdev, &cdemu_ctl_fops);
    kobject_set_name(&dev->cdev.kobj, "cdemu_ctl%i", which);
    result = cdev_add(&dev->cdev, devno, 1);
    
    dev->ctl_opened = 0;
    
    init_waitqueue_head (&dev->read_wait);
    init_waitqueue_head (&dev->write_wait);    
    
    mutex_init(&dev->req_mutex);
    
    dev->stage = STAGE_DEVICE_READY;
    dev->debug_mask = 0; /* No debugging by default */
    
    if (result) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to add character device cdemu_ctl%d (result: %i)\n", __func__, which, result);
    }
    
    /* Device's number and name... */
    dev->number = which;
    sprintf(dev->name, "cdemu%i", which);
    
    dev->driver = &cdemu_driver;
    
    /* Add control device to control devices' class */
    dev->ctl_class_device = class_device_create(cdemu_driver.cdemu_ctl_class, NULL, devno, NULL, "cdemu_ctl%d", which);
    
    /* Linux device model */
    dev_set_drvdata(&dev->ldm_device, dev);
    dev->ldm_device.bus = &cdemu_bus_type;
    dev->ldm_device.driver = &cdemu_driver.driver; /* Force driver binding: i.e. don't call match() on bus */
    dev->ldm_device.parent = &cdemu_busdev;
    dev->ldm_device.release = cdemu_device_release;
    strcpy(dev->ldm_device.bus_id, dev->name);
    result = device_register(&dev->ldm_device);
    if (result) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to register device!\n", __func__);
    }
    
    /* Attribute: stage */
    result = device_create_file(&dev->ldm_device, &dev_attr_stage);
    if (result) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to create 'stage' attribute!\n", __func__);
        return result;
    }
    /* Attribute: debug_mask */
    result = device_create_file(&dev->ldm_device, &dev_attr_debug_mask);
    if (result) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to create 'debug_mask' attribute!\n", __func__);
        return result;
    }
    
    /* Block device init */
    spin_lock_init(&dev->lock);
    init_waitqueue_head(&dev->request_wait);

    dev->queue = blk_init_queue(cdemu_block_request, &dev->lock);
    if (dev->queue == NULL) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to create request queue!\n", __func__);
        return -ENOMEM;
    }
    
    dev->queue->queuedata = dev;
    blk_queue_prep_rq(dev->queue, cdemu_block_prep_rq);
    
    /* CD-ROM sectors are 2048-byte ones... */
    blk_queue_hardsect_size(dev->queue, CD_FRAMESIZE); 
    /* Some other things IDE CD-ROM driver sets... */
    blk_queue_dma_alignment(dev->queue, 31);
    dev->queue->unplug_delay = (1 * HZ) / 1000;
    if (!dev->queue->unplug_delay) {
        dev->queue->unplug_delay = 1;
    }
    /* We play safe and disable readahead which sometimes causes locks... and I
       don't really think it's helping us any, anyway */
    dev->queue->backing_dev_info.ra_pages = 0;

    dev->gd = alloc_disk(1);
    if (!dev->gd) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to create gendisk!\n", __func__);
        return -ENOMEM;
    }
    dev->gd->major = param_block_major;
    dev->gd->first_minor = which;
    dev->gd->minors = 1;
    dev->gd->fops = &cdemu_block_ops;
    dev->gd->flags = GENHD_FL_CD | GENHD_FL_REMOVABLE;
    dev->gd->queue = dev->queue;
    dev->gd->private_data = dev;
    dev->gd->driverfs_dev = &dev->ldm_device;
    strcpy(dev->gd->disk_name, dev->name);
    
    /* add_disk() should be called *after* register_cdrom()... this is because 
       block device's open calls cdrom_open(), and an oops ensues if we aren't 
       registered with uniform cd-rom driver yet */
    
    /* CD-ROM device init */
    dev->cdi.ops = &cdemu_cdrom_dops;
    dev->cdi.speed = 40; /* Max speed */
    dev->cdi.capacity = 1; /* Number of discs in jukebox */
    dev->cdi.handle = dev;
    strcpy(dev->cdi.name, dev->name);
    dev->cdi.disk = dev->gd;
    if (register_cdrom(&dev->cdi)) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to register with uniform CD-ROM driver!\n", __func__);
    }
    
    /* Finish the registration; look the comment at the end of 'Block device 
       init' if you really want to know why this is here and not there */
    add_disk(dev->gd);
    
    return 0;
}

static void cdemu_unsetup_device (struct cdemu_device *dev, int which) {
    /* Character control device */
    cdev_del(&dev->cdev);
    
    /* Remove control device from control devices' class */
    class_device_del(dev->ctl_class_device);
    
    /* Block device */
    if (dev->gd) {
        del_gendisk(dev->gd);
        put_disk(dev->gd);
    }
    if (dev->queue) {
        blk_cleanup_queue(dev->queue);
    }
    
    /* CD-ROM device */
    if (unregister_cdrom(&dev->cdi)) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to unregister with uniform CD-ROM driver!\n", __func__);
    }

    /* Delete attributes */
    device_remove_file(&dev->ldm_device, &dev_attr_stage);
    device_remove_file(&dev->ldm_device, &dev_attr_debug_mask);
    
    /* Unregister device */
    device_unregister(&dev->ldm_device);
    
    return;
}

static int __init cdemu_init (void) {
    int result = 0;
    int i = 0;
    
    /* Create bus */
    result = cdemu_bus_create();
    if (result) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to create 'cdemu-bus'!\n", __func__);
        return result;
    }
    
    /* Create driver */
    result = cdemu_driver_create();
    if (result) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to create 'cdemu-driver'!\n", __func__);
        return result;
    }
    
    /* Allocate the device array and initialize each device */
    devices = kmalloc(param_devices * sizeof(struct cdemu_device), GFP_KERNEL);
    if (devices == NULL) {
        CDEMU_DEBUG_DRV(MODULE_DEBUG_ERROR, "%s: failed to allocate devices!\n", __func__);    
        return -ENOMEM;
    }
    
    for (i = 0; i < param_devices; i++) {
        cdemu_setup_device(devices + i, i);
    }
     
    return 0;
}

static void cdemu_exit (void) {
    int i;
    /* Destroy devices */
    for (i = 0; i < param_devices; i++) {
        cdemu_unsetup_device(devices + i, i);
    }
    
    /* Free devices */
    kfree(devices);
    
    /* Destroy driver */
    cdemu_driver_destroy();
    
    /* Destroy bus */
    cdemu_bus_destroy();
}

module_init(cdemu_init);
module_exit(cdemu_exit);
